//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by resource.rc
//
#define IDD_MYRUNAS                     101
#define IDC_RUNASCURR                   1000
#define IDC_RUNASSPEC                   1001
#define IDC_SHICON                      1002
#define IDC_HELPTEXT                    1003
#define IDC_USERNAME                    1004
#define IDC_PASSWORD                    1005
#define IDC_LBLUSER                     1007
#define IDC_LBLPWD                      1008

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        102
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1009
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
