//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"
#include "uac.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_MYRUNAS DIALOG DISCARDABLE  0, 0, 250, 145
STYLE DS_MODALFRAME | DS_NOIDLEMSG | DS_SETFOREGROUND | DS_FIXEDSYS | 
    DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
FONT 8, "MS Shell Dlg"
BEGIN
    DEFPUSHBUTTON   "&OK",IDOK,132,122,50,14
    PUSHBUTTON      "Ca&ncel",IDCANCEL,188,122,50,14
    ICON            "",IDC_SHICON,7,7,20,20
    LTEXT           "",IDC_HELPTEXT,34,7,204,35
    CONTROL         "",IDC_RUNASCURR,"Button",BS_AUTORADIOBUTTON,20,49,218,
                    10
    CONTROL         "",IDC_RUNASSPEC,"Button",BS_AUTORADIOBUTTON,20,65,218,
                    10
    LTEXT           "&User name:",IDC_LBLUSER,20,84,42,16
    EDITTEXT        IDC_USERNAME,63,83,175,14,ES_AUTOHSCROLL
    LTEXT           "&Password:",IDC_LBLPWD,20,102,42,20
    EDITTEXT        IDC_PASSWORD,63,100,175,14,ES_PASSWORD | ES_AUTOHSCROLL
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_MYRUNAS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 238
        TOPMARGIN, 7
        BOTTOMMARGIN, 136
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""afxres.h""\r\n"
    "#include ""uac.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

